# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""AWS configuration."""

VENDOR_CODE: str = "deadline"

# S3 related
S3_CONNECT_TIMEOUT_IN_SECS: int = 30
S3_READ_TIMEOUT_IN_SECS: int = 30
S3_RETRIES_MODE: str = "standard"
